/****************************************************************
 *  KLINGEL.C							*
 *			FlashEEPROM-Klingel			*
 *			===================			*
 *								*
 *  05.02.96	Bodo Wenzel	V1.0, Erstellung		*
 *  08.02.96	Bodo Wenzel	V1.1, Verkrzung auf unter 4KB	*
 *  20.02.96	Bodo Wenzel	V1.2, Latch fr DAC		*
 *  30.09.96	Bodo Wenzel	V1.3, kleine Verbesserungen	*
 *  21.07.98	Bodo Wenzel	V1.4, Anzeigefehler beseitigt	*
 ****************************************************************/

#define	VERSION	"V1.4"

/* Aufruf mit "define(TEST)", wenn Testversion erstellt wird!	*/

/* === Einfgedateien ========================================= */

#include <ctype.h>
#include <intrins.h>
#include <string.h>

#ifdef TEST
#include <reg515.h>
#else
#include <reg51.h>
#endif

/* === Konstanten ============================================= */

#define	FALSE	0
#define	TRUE	(!FALSE)

#define	QUARZ	11059200L	/* Quarzfrequenz in Hz		*/
#define	TICKS	8000		/* Samplefrequenz in Hz		*/
#define	BAUD	9600		/* Baudrate			*/
#define	BSIZE	128		/* Blockgre			*/

#define	SOH	0x01		/* Steuerzeichen		*/
#define	EOT	0x04
#define	ACK	0x06
#define	BS	0x08
#define	CR	0x0D
#define	XON	0x11
#define	XOFF	0x13
#define	NAK	0x15
#define	CAN	0x18
#define	EOF	(-1)		/* kein Zeichen			*/

enum {	ERR_NONE=0,		/* Fehlercodes			*/
	ERR_CHAR, ERR_BLOCK, ERR_CHECKSUM, ERR_TIMEOUT,
	ERR_ERASE, ERR_PROGRAM,
	ERR_FILLED, ERR_FULL };

#define	ML_TIMEOUT	30000U	/* Wartezeit in Hauptschleife	*/

#define	COM_LEN		32	/* Lnge des Kommentars		*/

/* === Typvereinbarungen ====================================== */

typedef	unsigned char	uchar;
typedef	unsigned int	uint;

typedef	struct {
  uchar		mode;		/* Modus: 0xFF, 'S', 'L'	*/
  uchar		ba_h,ba_m,ba_l;	/* Startadresse im Flash	*/
  uchar		ea_h,ea_m,ea_l;	/* Endadresse im Flash		*/
  char		com[COM_LEN];	/* Kommentar			*/
  }			header;

/* === Globale Variablen ====================================== */

uchar		error;		/* letzte Fehlernummer		*/

uchar		flash_cnt;	/* Zhler fr Versuche		*/

header	idata	sound;		/* aktueller Ton		*/
bit		sound_down;	/* Ton fr gedrckten Knopf	*/
bit		sound_off;	/* Ton aus			*/
uchar	code	*sound_rom;	/* Startadresse im ROM		*/
uint		sound_len;	/* Tonlnge in Blcken		*/

#ifdef TEST
uchar		sp;		/* maximaler Stackpointer	*/
#define	SP_GAP	4		/* zustzlich ntiger Stack	*/
#endif

/* === Input / Output ========================================= */

sbit	POWER	=P3^5;		/* Stromversorgung		*/

sbit	BUTTON	=P3^6;		/* Klingelknopf			*/

sbit	F_WE	=P3^3;		/* Flash-Schreibsignal		*/
sbit	F_OE	=P3^4;		/* Flash-Lesesignal		*/

#define	IO_D	P1		/* Datenbus			*/

sbit	AD_H	=P3^2;		/* Adrebus			*/
#ifdef TEST
#define	AD_M	P5
#define	AD_L	P4
#else
#define	AD_M	P2
#define	AD_L	P0
#endif

sbit	DAC	=P3^7;		/* Latchsignal fr DAC		*/

/* === Funktionsprototypen ==================================== */

bit button_down(void);
bit button_up(void);

void xm_upload(void);
void xm_download(void);

void flash_init(void);
void flash_erase(void);
void flash_inc(void);
bit flash_write(uchar d);
uchar flash_read(void);

void sound_setup(void);
uchar sound_tick(void);

void ser_init(void);
char ser_putc(char c);
void ser_puts(char *s);
void ser_puti(uint i);
int ser_getc(uint t);
void ser_gets(char idata *s,uchar n);

void time_init(void);
void time_wait(uchar t);

/* === M A I N ================================================ */

void main(void) {
/*-------------*/
char	c;

POWER=TRUE;			/* sicher einschalten		*/

sound_off=TRUE;
ser_init();
time_init();
flash_init();			/* Initialisierungen		*/

#ifdef TEST
EAL=TRUE;			/* Interrupts freigeben		*/
#else
EA=TRUE;
#endif

for (;;) {
  if (button_down())
    break;			/* Tne spielen, aber		*/
  if (button_up())		/* bei Hostanschlu abbrechen	*/
    break;
  if (BUTTON==1) {		/* Klingel noch losgelassen	*/
    POWER=FALSE;		/* ausschalten			*/
    for (;;);
    }
  }

sound_off=TRUE;
sound_down=TRUE;
for (;;) {
  sound_setup();		/* aktuellen Klang setzen	*/

  ser_puts("\r\nChip's FlashEEPROM Bell " VERSION "/" __DATE__);
#ifdef TEST
  ser_puts(" SP=");
  ser_puti(sp+SP_GAP);
#endif
  ser_puts("\r\nbutton=");
  ser_puts(sound_down ? "DOWN" : "UP  ");
  ser_puts(": mode='");
  ser_putc(sound.mode==0xFF ? 's' : sound.mode);
  ser_puts("', ");
  ser_puti(sound_len);
  ser_puts(" block(s), comment=\"");
  ser_puts(sound.com);
  ser_puts("\"\r\n");

  c=toupper(ser_getc(ML_TIMEOUT));	/* Zeichen vom Host	*/

  error=ERR_NONE;		/* bisher kein Fehler		*/

  switch (c) {			/* Kommandos ausfhren		*/
  case 'E':
    flash_erase();
    break;
  case 'D':
    sound_down=TRUE;
    break;
  case 'U':
    sound_down=FALSE;
    break;
  case 'S':
    sound_off=FALSE;
    while (!sound_off && !RI);
    sound_off=TRUE;
    break;
  case 'W':
    xm_upload();
    break;
  case 'R':
    xm_download();
    break;
  case EOF:
  case 'O':
    ser_puts("Switching off...");
    POWER=FALSE;
    for (;;);
    break;
  default:			/* bei Bedarf Hilfe anzeigen	*/
    ser_puts("Available commands:\r\n"
             "E  erase whole EEPROM\r\n"
             "D  select button down\r\n"
             "U  select button up\r\n"
             "S  start playback\r\n"
             "W  upload and write sample (Xmodem)\r\n"
             "R  read and download sample (Xmodem)\r\n"
             "O  power off\r\n");
    break;
    }

  switch (error) {		/* Fehler melden		*/
  case ERR_NONE:
    break;
  case ERR_CHAR:
    ser_puts("Wrong character!\r\n");
    break;
  case ERR_BLOCK:
    ser_puts("Wrong block number!\r\n");
    break;
  case ERR_CHECKSUM:
    ser_puts("Wrong checksum!\r\n");
    break;
  case ERR_TIMEOUT:
    ser_puts("Timeout!\r\n");
    break;
  case ERR_ERASE:
    ser_puts("Flash didn't erase!\r\n");
    break;
  case ERR_PROGRAM:
    ser_puts("Flash didn't program!\r\n");
    break;
  case ERR_FILLED:
    ser_puts("Sample not empty!\r\n");
    break;
  case ERR_FULL:
    ser_puts("Memory overflow!\r\n");
    break;
  default:
    ser_puts("Unknown error ");
    ser_puti(error);
    ser_puts("!\r\n");
    break;
    }
  }
}

bit button_down(void) {
/*-------------------*/
sound_down=TRUE;
sound_setup();
sound_off=FALSE;		/* Ton (gedrckt) spielen	*/

while (BUTTON==0)		/* solange Klingel		*/
  if (ser_getc(1)==CR)
    return(TRUE);		/* Host melden			*/

sound_off=TRUE;
return(FALSE);
}

bit button_up(void) {
/*-----------------*/
sound_down=FALSE;
sound_setup();
sound_off=FALSE;		/* Ton (losgelassen) spielen	*/

while (BUTTON==1 && !sound_off)	/* solange Klingel und Ton	*/
  if (ser_getc(1)==CR)
    return(TRUE);		/* Host melden			*/

sound_off=TRUE;
return(FALSE);
}

/* === XModem-Transfer ======================================== */
/* eingeschrnkte Implementation, die von fehlerfreier		*/
/* Kommunikation ausgeht...					*/

#define	XM_RETRY	10	/* Anzahl Versuche		*/

#define	XM_TX_TIME	60000U	/* Sendetimeout (Start)		*/
#define	XM_RX_TIME	10000U	/* Empfangstimeout (Start)	*/
#define	XM_C_TIME	1000U	/* Empfangstimeout (Zeichen)	*/

bit xm_r_block(uchar nr);
void xm_purge(void);

void xm_upload(void) {
/*------------------*/
uchar	i;
uchar	idata	*p;

if (sound.mode!=0xFF) {		/* Sample bereits programmiert	*/
  error=ERR_FILLED;
  return;
  }

ser_puts("Please start upload...\r\n");

i=0;
for (;;) {
  ser_putc(NAK);		/* Sendeaufforderung		*/

  if (ser_getc(XM_RX_TIME)==SOH)	/* Start abwarten	*/
    break;

  if (++i==XM_RETRY) {		/* Anzahl Versuche begrenzen	*/
    error=ERR_TIMEOUT;
    return;
    }
  }

flash_cnt=0;
sound_len=0;
i=0;
while (xm_r_block(++i));	/* Blcke empfangen		*/
ser_puti(sound_len);
ser_puts(" block(s) received.\r\n");

if (error!=ERR_NONE) {		/* bei Fehler abbrechen		*/
  ser_puts("Please erase, error: ");
  return;
  }

ser_puts("comment (");
ser_puti(COM_LEN-1);
ser_puts(" chars max):");
ser_gets(sound.com,COM_LEN-1);	/* weitere Daten abfragen	*/
if (!sound_down) {
  ser_puts("mode always Single!\r\n");
  i='S';
} else
  for (;;) {
    ser_puts("mode Single or Loop:");
    i=ser_putc(toupper(ser_getc(0)));
    ser_puts("\r\n");
    if (i=='S' || i=='L')
      break;
    }
sound.mode=i;
sound.ea_h=AD_H;		/* Endadresse merken		*/
sound.ea_m=AD_M;
sound.ea_l=AD_L;

AD_H=0;				/* Header speichern		*/
AD_M=0;
AD_L=sound_down ? 0 : sizeof(sound);
p=(uchar idata *)(&sound);
for (i=0; i<sizeof(sound); i++) {
  if (!flash_write(*p++))
    return;
  flash_inc();
  }
ser_puti(flash_cnt);
ser_puts(" write(s) max.\r\n");
}

void xm_download(void) {
/*--------------------*/
uint	n;
uchar	i,j,cs;

ser_puts("Please start download...\r\n");

switch (ser_getc(XM_TX_TIME)) {	/* Sendeaufforderung abwarten	*/
case NAK:
  break;
case EOF:
  error=ERR_TIMEOUT;
  return;
default:
  error=ERR_CHAR;
  return;
  }

i=0;
for (n=0; n<sound_len; n++) {	/* Blcke senden		*/
  ser_putc(SOH);
  ser_putc(++i);
  ser_putc(~i);
  cs=0;
  for (j=0; j!=BSIZE; j++)
    cs+=ser_putc(sound_tick());
  ser_putc(cs);

  switch (ser_getc(XM_C_TIME)) {	/* Quittierung berpr.	*/
  case ACK:
    break;
  case EOF:
    error=ERR_TIMEOUT;
    return;
  default:
    error=ERR_CHAR;
    return;
    }
  }

ser_putc(EOT);			/* Dateiende			*/
switch (ser_getc(XM_C_TIME)) {
case ACK:
  break;
case EOF:
  error=ERR_TIMEOUT;
  break;
default:
  error=ERR_CHAR;
  break;
  }
}

bit xm_r_block(uchar nr) {
/*----------------------*/
int	c;
uchar	i,cs;

if (ser_getc(XM_C_TIME)!=(int)nr) {	/* Blocknummer		*/
  xm_purge();
  error=ERR_BLOCK;
  return(FALSE);
  }
if (ser_getc(XM_C_TIME)!=(int)(~nr)) {	/* kompl. Blocknummer	*/
  xm_purge();
  error=ERR_BLOCK;
  return(FALSE);
  }

cs=0;
for (i=0; i!=BSIZE; i++) {	/* ein Block Daten		*/
  c=ser_getc(XM_C_TIME);
  if (c==EOF) {			/* Zeichen holen		*/
    xm_purge();
    error=ERR_TIMEOUT;
    return(FALSE);
    }
  if (!flash_write((uchar)c)) {	/* sofort speichern		*/
    xm_purge();
    return(FALSE);
    }
  flash_inc();
  if (AD_H==0 && AD_M==0 && AD_L==0) {	/* Lnge berwachen	*/
    xm_purge();
    error=ERR_FULL;
    return(FALSE);
    }

  cs+=(uchar)c;
  }

if (ser_getc(XM_C_TIME)!=(int)cs) {	/* Prfsumme		*/
  xm_purge();
  error=ERR_CHECKSUM;
  return(FALSE);
  }

ser_putc(ACK);			/* Empfang quittieren		*/
sound_len++;

switch (ser_getc(XM_C_TIME)) {	/* nchstes Zeichen		*/
case SOH:			/* weiterer Block		*/
  return(TRUE);
case EOT:			/* Dateiende			*/
  ser_putc(ACK);
  return(FALSE);
default:			/* Fehler			*/
  xm_purge();
  error=ERR_TIMEOUT;
  return(FALSE);
  }
}

void xm_purge(void) {
/*-----------------*/
while (ser_getc(XM_C_TIME)!=EOF);	/* Sendepause abwarten	*/

ser_putc(CAN);			/* Sendung abbrechen		*/
}

/* === EEPROM ================================================= */
/* !!! ACHTUNG !!! Die Funktionen gehen von einem Quarz mit	*/
/* ca. 10..12 MHz aus!						*/

#if QUARZ<10000000L
#error QUARZ zu klein!
#elif QUARZ>12000000L
#error QUARZ zu gro!
#endif

#define	FLASH_E_RETRY	1000U	/* Anzahl Lschversuche		*/
#define	FLASH_P_RETRY	25	/* Anzahl Programmierversuche	*/

void flash_init(void) {
/*-------------------*/
AD_H=0;
AD_M=0;
AD_L=0;				/* Adresse setzen		*/

IO_D=0xFF;			/* "reset"			*/
F_WE=0;
F_WE=1;
IO_D=0xFF;			/* "reset"			*/
F_WE=0;
F_WE=1;
IO_D=0x00;			/* "read mode"			*/
F_WE=0;
F_WE=1;
}

void flash_erase(void) {
/*--------------------*/
uint	i;
uchar	r;

ser_puts("Programming all bytes to 0...\r\n");

AD_H=0;
AD_M=0;
AD_L=0;
flash_cnt=0;
do {				/* Flash zuerst auf 0 setzen	*/
  if (!flash_write(0x00))
    return;
  flash_inc();
  if ((AD_M&0x0F)==0 && AD_L==0)
    ser_putc('.');
} while (AD_L!=0 || AD_M!=0 || AD_H!=0);
ser_puti(flash_cnt);
ser_puts(" write(s) max.\r\n"
         "Erasing...\r\n");

AD_H=0;
AD_M=0;
AD_L=0;
for (i=0; i<FLASH_E_RETRY; i++) {
  IO_D=0x20;			/* "setup"			*/
  F_WE=0;
  F_WE=1;
  IO_D=0x20;			/* "erase"			*/
  F_WE=0;
  F_WE=1;
  time_wait(TICKS/100);		/* Standby...			*/

  do {
    IO_D=0xA0;			/* "verify"			*/
    F_WE=0;
    F_WE=1;
    _nop_(); _nop_(); _nop_();	/* Standby...			*/
    _nop_(); _nop_(); _nop_();
    IO_D=0xFF;
    F_OE=0;
    r=IO_D;			/* Daten zurcklesen		*/
    F_OE=1;
    if (r!=0xFF)
      break;
    flash_inc();
  } while (AD_L!=0x00 || AD_M!=0x00 || AD_H!=0);

  if (r==0xFF)			/* komplett gelscht		*/
    break;
  }

IO_D=0xFF;			/* "reset"			*/
F_WE=0;
F_WE=1;
IO_D=0xFF;			/* "reset"			*/
F_WE=0;
F_WE=1;
IO_D=0x00;			/* "read mode"			*/
F_WE=0;
F_WE=1;
if (r!=0xFF)			/* Fehler melden		*/
  error=ERR_ERASE;

ser_puti(++i);
ser_puts(" pulse(s).\r\n");
}

void flash_inc(void) {
/*------------------*/
if (++AD_L==0x00)		/* Ports direkt hochzhlen	*/
  if (++AD_M==0x00)
    AD_H=~AD_H;
}

bit flash_write(uchar d) {
/*----------------------*/
uchar	i,r;

r=flash_read();
if (r==d)			/* Daten bereits vorhanden	*/
  return(TRUE);
if ((d&(r^d))!=0x00) {		/* Daten nicht programmierbar	*/
  error=ERR_PROGRAM;
  return(FALSE);
  }

for (i=0; i<FLASH_P_RETRY; i++) {
  IO_D=0x40;			/* "setup"			*/
  F_WE=0;
  F_WE=1;
  IO_D=d;			/* Daten schreiben		*/
  F_WE=0;
  F_WE=1;
  _nop_(); _nop_(); _nop_();	/* Standby...			*/
  _nop_(); _nop_(); _nop_(); _nop_(); _nop_(); _nop_(); _nop_();

  IO_D=0xC0;			/* "verify"			*/
  F_WE=0;
  F_WE=1;
  _nop_(); _nop_(); _nop_();	/* Standby...			*/
  _nop_(); _nop_(); _nop_();
  IO_D=0xFF;
  F_OE=0;
  r=IO_D;			/* Daten zurcklesen		*/
  F_OE=1;

  if (r==d)			/* Daten OK			*/
    break;
  }
if (flash_cnt<++i)		/* Maximum merken		*/
  flash_cnt=i;

IO_D=0xFF;			/* "reset"			*/
F_WE=0;
F_WE=1;
IO_D=0xFF;			/* "reset"			*/
F_WE=0;
F_WE=1;
IO_D=0x00;			/* "read mode"			*/
F_WE=0;
F_WE=1;
if (r!=d) {			/* Fehler melden		*/
  error=ERR_PROGRAM;
  return(FALSE);
  }
return(TRUE);
}

uchar flash_read(void) {
/*--------------------*/
IO_D=0xFF;
F_OE=0;
IO_D=IO_D;			/* Port hlt Daten fest		*/
F_OE=1;
return(IO_D);
}

/* === Klingel ================================================ */

/*	Tontabelle: Werte.., 0, Lnge-1				*/

uchar	code	sound_high[]={	/* hoher Sinus			*/
	  48, 89,117,127,117, 89, 48,
	  0, 8 };

uchar	code	sound_low[]={	/* tiefer Sinus			*/
	  39, 74,102,120,127,120,102, 74, 39,
	  0, 10 };

uchar	code	sound_exp[]={	/* Abschwellzeiten		*/
  6,  6,  6,  7,  6,  7,  6,  7,  6,  7,  6,  7,  6,  7,  6,  7,
  7,  7,  6,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,
  7,  7,  8,  7,  7,  8,  7,  8,  7,  7,  8,  8,  7,  8,  7,  8,
  8,  8,  8,  7,  8,  8,  8,  8,  8,  8,  9,  8,  8,  8,  9,  8,
  8,  9,  8,  9,  8,  9,  9,  8,  9,  9,  9,  9,  9,  9,  9,  9,
  9,  9, 10,  9,  9, 10,  9, 10,  9, 10, 10, 10,  9, 10, 10, 10,
 10, 11, 10, 10, 10, 11, 10, 11, 11, 10, 11, 11, 11, 11, 11, 11,
 11, 12, 11, 11, 12, 12, 11, 12, 12, 12, 12, 12, 12, 13, 12, 13,
 12, 13, 13, 13, 13, 13, 14, 13, 13, 14, 14, 14, 14, 14, 14, 14,
 15, 14, 15, 15, 15, 15, 15, 16, 16, 15, 16, 16, 17, 16, 17, 16,
 17, 17, 18, 17, 18, 18, 18, 18, 19, 18, 19, 19, 20, 19, 20, 20,
 21, 20, 21, 22, 21, 22, 22, 22, 23, 23, 23, 24, 24, 25, 24, 26,
 25, 26, 27, 26, 28, 28, 28, 29, 29, 30, 31, 31, 31, 33, 33, 33,
 35, 35, 36, 37, 37, 39, 39, 41, 41, 43, 44, 45, 46, 48, 49, 51,
 53, 54, 56, 58, 61, 62, 66, 68, 71, 74, 78, 82, 87, 91, 97,104,
110,119,128,139,152,169,188,214,247,255,255,255,255,255,255,255 };

uchar		sound_e,sound_i,sound_t;
bit		sound_s;

void sound_setup(void) {
/*--------------------*/
uchar	i;
uchar	idata	*p;

AD_H=0;				/* ersten Header lesen		*/
AD_M=0;
AD_L=sound_down ? 0 : sizeof(sound);
p=(uchar idata *)(&sound);
for (i=0; i<sizeof(sound); i++) {
  *p++=flash_read();
  flash_inc();
  }

if (sound.mode==0xFF) {		/* wenn nicht programmiert,	*/
  AD_H=0;			/* zweiten Header lesen		*/
  AD_M=0;
  AD_L=sound_down ? sizeof(sound) : 0;
  p=(uchar idata *)(&sound);
  for (i=0; i<sizeof(sound); i++) {
    *p++=flash_read();
    flash_inc();
    }

  if (sound.mode==0xFF) {	/* erste freie Adresse setzen	*/
    sound.ba_h=0;
    sound.ba_m=0;
    sound.ba_l=2*sizeof(sound);
  } else {
    sound.ba_h=sound.ea_h;
    sound.ba_m=sound.ea_m;
    sound.ba_l=sound.ea_l;
    }

  sound.mode=0xFF;		/* Daten korrigieren		*/
  sound_len=sound_exp[0];
  for (i=1; i!=0; i++)
    sound_len+=sound_exp[i];
  sound_len+=BSIZE-1;
  sound_len/=BSIZE;
  strcpy(sound.com,sound_down ? "Ding (ROM)" : "Dong (ROM)");
  sound_rom=sound_down ? sound_high : sound_low;

  sound_s=FALSE;		/* Abspielen vorbereiten	*/
  sound_e=255;
  sound_i=0;
  sound_t=sound_exp[sound_i];

} else {			/* sonst Lnge berechnen	*/
  sound_len=(uint)(sound.ea_h)<<8;
  sound_len+=sound.ea_m;
  sound_len-=(uint)(sound.ba_h)<<8;
  sound_len-=sound.ba_m;
  sound_len*=2;
  if (sound.ea_l&0x80)
    sound_len++;
  if (sound.ba_l&0x80)
    sound_len--;
  }

AD_H=sound.ba_h;		/* Startadresse setzen		*/
AD_M=sound.ba_m;
AD_L=sound.ba_l;

IO_D=0x80;			/* Ruhepegel einstellen		*/
DAC=0;
DAC=1;
}

uchar sound_tick(void) {
/*--------------------*/
uchar	v;

if (sound.mode==0xFF) {		/* Ton aus ROM			*/

  if (sound_s)			/* je nach Halbwelle		*/
    v=0x80-((uint)(*sound_rom*sound_e)>>8);
  else
    v=0x80+((uint)(*sound_rom*sound_e)>>8);

  if (*sound_rom++==0) {	/* nchster Sinuswert		*/
    sound_rom-=*sound_rom;
    sound_s=~sound_s;
    }

  if (--sound_t==0) {		/* nchster Abschwellwert	*/
    if (sound_e!=0)
      sound_e--;
    else
      sound_off=TRUE;
    sound_t=sound_exp[++sound_i];
    }

} else {			/* Ton aus Flash		*/

  v=flash_read();		/* nchster Tonwert		*/
  flash_inc();
  if (AD_H==sound.ea_h &&	/* Ende berwachen		*/
      AD_M==sound.ea_m &&
      AD_L==sound.ea_l) {
    if (sound.mode=='L') {
      AD_H=sound.ba_h;
      AD_M=sound.ba_m;
      AD_L=sound.ba_l;
    } else
      sound_off=TRUE;
    }
  }

return(IO_D=v);
}

/* === Serielle Schnittstelle ================================= */

#define	TIMER1	(QUARZ/12/16/BAUD)

void ser_init(void) {
/*-----------------*/
SCON=0x58;			/* 8-N-1, Empfang freigeben	*/
PCON|=0x80;			/* doppelte Baudrate		*/

TMOD&=0x0F;
TMOD|=0x20;
TH1=-TIMER1;
TR1=TRUE;			/* Timer 1 als Baudratentimer	*/

TI=TRUE;
RI=FALSE;			/* Kommunikation vorbereiten	*/
}

char ser_putc(char c) {
/*-------------------*/
if (RI) {			/* XON/XOFF-Protokoll		*/
  if (SBUF==XOFF) {
    do {
      RI=FALSE;
      while (!RI);
    } while (SBUF!=XON);
    RI=FALSE;
    }
  }
while (!TI);			/* letztes Zeichen abwarten	*/
TI=FALSE;
return(SBUF=c);
}

void ser_puts(char *s) {
/*--------------------*/
while (*s)
  (void)ser_putc(*s++);
}

void ser_puti(uint i) {
/*-------------------*/
uchar	r;
uint	d;
bit	o;

o=FALSE;			/* bisher kein Zeichen		*/
for (d=10000; d!=1; d/=10) {
  r=(uchar)(i/d);
  if (r!=0)
    o=TRUE;			/* fhrende Nullen unterdrcken	*/
  if (o)
    ser_putc(r+'0');
  i-=r*d;			/* ausgegebene Ziffer "lschen"	*/
  }
ser_putc(i+'0');		/* mindestens eine Ziffer	*/
}

int ser_getc(uint t) {
/*------------------*/
uchar	i;

for (;;) {
  for (i=0; i!=TICKS/1000; i++) {	/* 1 ms warten		*/
    if (RI) {
      i=SBUF;
      RI=FALSE;
      return(i);
      }
    time_wait(1);
    }
  if (t!=0)			/* mit Timeout warten		*/
    if (--t==0)
      return(EOF);
  }
}

void ser_gets(char idata *s,uchar n) {
/*----------------------------------*/
uchar	i,c;

for (i=0;;)
  switch(c=ser_getc(0)) {	/* Zeichen holen		*/
  case BS:
    if (i!=0) {			/* letztes Zeichen lschen	*/
      ser_puts("\b \b");
      s--;
      i--;
      }
    break;
  case CR:
    ser_puts("\r\n");		/* Eingabe fertig		*/
    *s='\0';
    return;
  default:
    if (c>=' ' && i!=n) {	/* weiteres Zeichen		*/
      *s++=ser_putc(c);
      i++;
      }
    break;
    }
}

/* === System-Timer =========================================== */

#define	TIMER0	(QUARZ/12/TICKS)

uchar		time_t;		/* Verzgerungszhler		*/

void time_init(void) {
/*------------------*/
TMOD&=0xF0;
TMOD|=0x02;
TH0=-TIMER0;
TR0=TRUE;			/* Timer vorbereiten		*/

TF0=FALSE;
ET0=TRUE;			/* Interrupt freigeben		*/

#ifdef TEST
sp=SP;				/* ersten Wert setzen		*/
#endif
}

void time_wait(uchar t) {
/*---------------------*/
time_t=t;
while (time_t!=0);		/* Zeit abwarten		*/
}

void timer0(void) interrupt 1 {
/*---------------------------*/
#ifdef TEST
DAC=0;				/* Start der ISR		*/

if (SP>sp)			/* Stackpointer mitfhren	*/
  sp=SP;
#endif

if (time_t!=0)
  time_t--;			/* Wartezeit bearbeiten		*/

if (!sound_off) {
  (void)sound_tick();		/* Klingel bearbeiten		*/
  DAC=0;
  }
DAC=1;
}

/* === Ende =================================================== */
